#!/usr/bin/perl
#  categories.pl - functions to implement categories
#
#  Written by Curtis Olson.  Started September 29, 1994.
#
#  Copyright (C) 1994 - 1999  Curtis L. Olson  - curt@me.umn.edu
#  Portions Copyright 2000 Eugene Magnier - eugene@cfht.hawaii.edu
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# accounts.pl,v 0.11 2001/08/12 20:47:36 briank Exp

package CBB;

use strict;    # don't take no guff

# Global variables

# %CBB::ACCTS - an associative array of accounts
# @CBB::ACCOUNTS - a sorted list of account keys (for traversing the cat list)
# $CBB::sorted_accts - specifies whether the list in @CBB::CATKEYS is valid
# $CBB::account_name - current account name, used as key for %CBB::ACCTS

# initialize the transactions data structure

sub init_trans {
    # in: [account name]
    # out: result

    my( $name ) = @_;

    $CBB::sorted_keys = 0;
    $CBB::calced = 0;
    @CBB::KEYS = ();
    %CBB::TRANS = ();
    @CBB::XFER = ();
    $CBB::account_name = $name;

    return "ok"; 
}


# make a new account
sub make_acct {
    # in: key rootPath file desc
    # out: result
    
    my($name, $path, $file, $desc) = split(/\t/, $_[0]);
    my($pos, $short_name, $result);

    print DEBUG "Make account $name - $desc\n" if $CBB::debug;

    init_trans ( $name );
    $CBB::current_file = "$file";
    
    # add acct to group list
    add_acct ("$name\t$file\t$desc\t0.0\t0.0\t0.0");
    $result = save_accts ();
    if ($result eq "error") { return "error 1"; }

    # save the categories file before it gets toasted
    insert_cat("[$CBB::group.$name]\t$desc\t");
    $result = save_cats( $path );
    if ($result eq "error") { return "error 2"; }

    # create the file
    $result = save_trans("$CBB::current_file");
    if ($result eq "error") { return "error 3"; }

    return "ok";
}

# each category in the file should correspond to one in the category list
# if not, return an error
sub check_cats {

    my ($cat, $key, $junk);

    if ($CBB::sorted_keys == 0) { &sort_keys(); }

    foreach $key (@CBB::KEYS) {
        ($junk, $junk, $junk, $junk, $junk, $cat, $junk, $junk, $junk) = split(/\t/, $CBB::TRANS{$key});

	# print STDERR "$key - $cat\n";
	# for now, ignore splits -- fix this later
	if ($cat =~ m/^\|.*\|$/) { next; }

	if ($CBB::CATS{$cat} eq "") {
	    return "error";
	}
    }
    return "ok";
}

# list missing categories, ending with 'none'
sub list_missing_cats {

    my ($cat, $key, $junk);
    my (@miss, $value, $found);

    if ($CBB::sorted_keys == 0) { &sort_keys(); }

    foreach $key (@CBB::KEYS) {
        ($junk, $junk, $junk, $junk, $junk, $cat, $junk, $junk, $junk) = split(/\t/, $CBB::TRANS{$key});

	if ($cat =~ m/^\|.*\|$/) { next; }

	if ($CBB::CATS{$cat} eq "") {
	    $found = 0;
	    foreach $value (@miss) {
		if ($value eq $cat) { 
		    $found = 1; 
		    last;
		}
	    }
	    if (!$found) {
		@miss = (@miss, $cat);
	    }
	}
    }

    foreach $value (@miss) {
	print "$value\n";
    }
    return "none";
}

# determine the file type and call the correct load routine
sub load_trans {
    # in: account filename base 
    # out: result

    my($file) = @_;

    if ($CBB::cache) {
    	no strict 'vars';	# necessary for this special hack
    	no strict 'refs';

	# save current data to cache
	my($hname) = "ACC_" . "$CBB::group." . &file_basename($CBB::current_file);
	print DEBUG "$hname $CBB::current_file\n" if $CBB::debug;
	%$hname = %CBB::TRANS;

	# test if new table already in cache
	$hname = "ACC_" . "$CBB::group." . &file_basename($file);
	print DEBUG "$hname\n" if $CBB::debug;
	if (scalar (%$hname) ) {
		print DEBUG "$hname defined , load from cache\n" if $CBB::debug;

		$CBB::sorted_keys = 0;
		$CBB::calced = 0;
		
		%CBB::TRANS = %$hname;	# take values from the cache
		&calc_trans();

		$CBB::current_file = $file;
		return "ok";
	}
    }
    return &load_cbb_trans($file);
}


# load the data from a cbb file
sub load_cbb_trans {
    # in: account filename base
    # out: result

    my($file) = @_;
    my($version) = "";
    my($group) = "";
    my($balance) = 0.0;
    my($ebalance) = 0.0;
    my($cbalance) = 0.0;
    my($junk);
    my($name) = "";

    my $filename = "$CBB::path/$CBB::group/$file";
    print DEBUG "load_cbb_trans(): Loading the cbb format file: $CBB::path/$CBB::group/$file\n" if $CBB::debug;
    print DEBUG "load_cbb_trans(): file 4: $filename\n" if $CBB::debug;

    if ( $CBB::decrypt ne "" ) {
	open(LOAD, "$CBB::decrypt < $filename|") || return "error";
    } else {
    	open (LOAD, "<$filename") || return "error";
        print DEBUG "load_cbb_trans(): opened $filename\n" if $CBB::debug;
    }

    while ( <LOAD> ) {
	if ( m/^#/ ) {
	     # toss the comment (but first check for any goodies.)
	     if ( m/^#\s*CBB Version\s*=/ ) {
		  ($version) = /^#\s*CBB Version\s*=\s*(.*)$/;
	      }
	     if ( m/^#\s*GROUP\*=/ ) {
		  ($group) = /^#\s*GROUP\s*=\s*(.*)$/;
	      }
	     if ( m/^#\s*Current Balance\s*=/ ) {
		  ($balance) = /^#\s*Current Balance\s*=\s*(.*)$/;
	      }
	     if ( m/^#\s*Ending Balance\s*=/ ) {
		  ($ebalance) = /^#\s*Ending Balance\s*=\s*(.*)$/;
	      }
	     if ( m/^#\s*Cleared Balance\s*=/ ) {
		  ($cbalance) = /^#\s*Cleared Balance\s*=\s*(.*)$/;
	      }
	     if ( m/^#\s*Account Name\s*=/ ) {
		  ($name) = /^#\s*Account Name\s*=\s*(.*)$/;
                  &init_trans( $name );
	      }
	} else {
	    if ( $version eq "") {
		print DEBUG "no data file version, file encrypted ?" if $CBB::debug;
		close(LOAD);
		return "error";
	    }
	    chop;
            &create_trans($_);
	}
    }
    close(LOAD);

    &calc_trans();

    $CBB::current_file = $file;

    print DEBUG "loaded the cbb format file\n" if $CBB::debug;

    return "ok";

}

sub auto_save_trans {
    # in: filename
    # out: result

    $CBB::autosave = 1;
    
    my($result) = save_trans ($_[0]);

    $CBB::autosave = 0;

    return ($result);
}


# save all the precious data to a file
sub save_trans {
    # in: filename [new account name] [new group name]
    # out: result

    my ($file, $newName, $newGroup) = split( /\t/, $_[0] );
    my($auto_save_file, $key);
    my(@trans);
    my($file_exists) = 0;
    my($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, 
       $ctime, $blksize, $blocks);
    if( !defined $newName ) {
      $newName = $CBB::account_name;
    }
    if( !defined $newGroup ) {
      $newGroup = $CBB::group;
    }
    

    $CBB::saving_trans = 0;
    $file = "$CBB::path/$newGroup/$file";

    print DEBUG "Saving the cbb format file: $file\n" if $CBB::debug;

    if ($CBB::calced == 0) { &calc_trans(); }
    if ($CBB::sorted_keys == 0) { &sort_keys(); }

    if ( $CBB::encrypt ne "" ) {
	open(SAVE, "|$CBB::encrypt > $file.new") || return "error";
    } else {
    	open(SAVE, ">$file.new") || return "error";
    }

    # Print some header stuff
    printf (SAVE "# CBB-Account File\n");
    printf (SAVE "#\n");
    printf (SAVE "# CBB Version = $CBB::version_num\n");
    printf (SAVE "# GROUP = $newGroup\n");
    printf (SAVE "# Account Name = $newName\n");
    printf (SAVE "# Current Balance = %.2f\n", $CBB::BALS{Current});
    printf (SAVE "# Ending Balance = %.2f\n", $CBB::BALS{Amount});
    printf (SAVE "# Transaction Count = $CBB::BALS{Count}\n");
    printf (SAVE "# Cleared Balance = %.2f\n", $CBB::BALS{Xamount});
    printf (SAVE "# Cleared Txn Count = $CBB::BALS{Xcount}\n");
    printf (SAVE "# Date Saved = %s\n", &raw_date());
    printf (SAVE "# User = $CBB::user_name\n");
    printf (SAVE "#\n");
    printf (SAVE "# date  check  desc  debit  credit  cat  com  cleared\n");
    printf (SAVE "# ---------------------------------------------------\n");

    # Update CBB::ACCTS
    &updateAccountTable;

    foreach $key (@CBB::KEYS) {
	# strip off last total
	@trans = split(/\t/, $CBB::TRANS{$key});
	print SAVE join ("\t", @trans[0..7]) . "\n";
    }

    if ($CBB::debug) {
	foreach $key (@CBB::XFER) {
	    print DEBUG "$key\n";
	}
    }


    if ($CBB::autosave) { 
	foreach $key (@CBB::XFER) {
	    print SAVE "$key\n";
	}
	close(SAVE);
    } else {
	close(SAVE);
	
	# preserve file permissions if the file exists
	if ( -e $file ) {
	    $file_exists = 1;
	    ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, 
	     $ctime, $blksize, $blocks) = stat($file);
	    print DEBUG "file permissions = $mode\n" if $CBB::debug;
	    print DEBUG "file owner = $uid  group = $gid\n" if $CBB::debug;
	}
	
	if ( $file_exists ) {
	    unlink("$file.bak");
	    rename("$file", "$file.bak");
	}
	
	rename("$file.new", "$file");
	
	if ( $file_exists ) {
	    chown($uid, $gid, $file);
	    chmod($mode, $file);
	}
	if (!$CBB::saving_trans) { apply_transfers (); }
    }

    # this is poor - we have two places which define the autosave filname format
    if (!$CBB::autosave) {
	$auto_save_file = &file_dirname($file) . "#" . &file_basename($file) . "#";
	print DEBUG "auto_save_file = $auto_save_file\n" if $CBB::debug;
	if ( -e $auto_save_file ) {
	    unlink("$auto_save_file");
	    unlink("$auto_save_file.bak");
	}
    }

    return "ok";
}

sub apply_transfers {

    # at the start, the current account has been saved, so we are
    # safe to load in the 'to' accounts one-by-one.  we must re-load
    # the current account at the end to return to the starting state
    # globals like CBB:current_file, etc are affected by these ops, so
    # we save the needed values here.

    my($from_file, $from_cat, @from_xfer, $from_group);
    my($line, $junk, $key, $i, $name, $trans);
    my(%tolist, @Tolist, $to_cat);
    my($date, $check, $desc, $debit, $credit, $cat, $com, $cleared, $mode);

    $CBB::saving_trans = 1;

    # save the name of the current file
    $from_file = $CBB::current_file;
    $line = find_acct_file ($CBB::current_file);
    ($from_cat, $junk) = split (/\t/, $line);
    $from_cat = "[" . $CBB::group . "." . $from_cat . "]";
    @from_xfer = @CBB::XFER;
    $from_group = $CBB::group;
    
    if ($CBB::debug) {
	print DEBUG "apply_transfers(): file: $from_file, cat: $from_cat, xfer: @from_xfer\n";
	
	foreach $key (@CBB::XFER) {
	    print DEBUG "$key\n";
	}
    }

    # first, find all of the 'to' accounts
    %tolist = ();
    foreach $key (@from_xfer) {
	($date, $check, $desc, $debit, $credit, $cat, $com, $cleared, $mode) = split (/\t/, $key);
	if ( $tolist{$cat} == "" ) { $tolist{$cat} = 1; }
    }
    
    # load each file, apply the appropriate transfers
    @Tolist = %tolist;
    for ($i = 0; $i < @Tolist; $i+=2) {
	$to_cat = $Tolist[$i];
	$name = $to_cat; 
        $name =~ s/\[(.*)\]/\1/;
        my( $to_group, $to_name ) = split( /\./, $name );
        if( $to_group ne $CBB::group ) {
          $CBB::group = $to_group;
        }
	print DEBUG "apply_transfers(): to: $to_cat, group: $to_group name: $to_name\n" if $CBB::debug;
    
	my $result = load_trans ($to_name);
        # Probably want some kind of user notification here...
        if( $result ne "ok" ) {
          if( $from_group ne $CBB::group ) {
            $CBB::group = $from_group;
          }
          print DEBUG "apply_transfers(): no such account: $to_cat\n" if $CBB::debug;
          next;
        }
	
	foreach $key (@from_xfer) {
	    print DEBUG "apply_transfers(): key: $key\n" if $CBB::debug;
	    ($date, $check, $desc, $debit, $credit, $cat, $com, $cleared, $mode) = split (/\t/, $key);
	    if ( $to_cat eq $cat ) { 
		$trans = "$date\t$check\t$desc\t$debit\t$credit\t$from_cat\t$com\t$cleared\t0.00";
		if ($mode eq "+") {
		    create_trans ($trans);
		}
		if ($mode eq "-") {
		    $key = find_trans_key ($trans);
		    delete_trans ($key);
		}
	    }
	}
        $CBB::calced = 0;
	save_trans ($to_name);
        &save_accts
    }

    # re-load original file
    print DEBUG "apply_transfers(): reloading original file\n" if $CBB::debug;
    $CBB::group = $from_group;
    $line = load_trans ($from_file);

    $CBB::saving_trans = 0;

}

sub updateAccountTable {
  # Populate some fields in %CBB::ACCTS
  if( exists $CBB::ACCTS{$CBB::account_name} ) {
    my( $file, $desc, $curbal, $endbal, $clearbal ) =
      split( /\t/, $CBB::ACCTS{$CBB::account_name} ); 
    $curbal = $CBB::BALS{Current};
    $endbal = $CBB::BALS{Amount};
    $clearbal = $CBB::BALS{Xamount};
    $CBB::ACCTS{$CBB::account_name} = 
      join( "\t", $file, $desc, $curbal, $endbal, $clearbal );
  }
}

1;

