#!/usr/bin/perl
#  categories.pl - functions to implement categories
#
#  Written by Curtis Olson.  Started September 29, 1994.
#
#  Copyright (C) 1994 - 1999  Curtis L. Olson  - curt@me.umn.edu
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# categories.pl,v 0.10 2001/08/12 20:47:36 briank Exp

package CBB;

use strict;    # don't take no guff

# Global variables

# $CBB::categories - category file for this group
# $CBB::group - group name
# %CBB::CATS - an associative array of categories
# @CBB::CATKEYS - a sorted list of category keys (for traversing the cat list)
# $CBB::sorted_catkeys - specifies whether the list in @CBB::CATKEYS is valid

# initialize the categories list
sub init_cats {
    # out: result

    print DEBUG "init_cats()\n" if $CBB::debug;

    %CBB::CATS = ();
    @CBB::CATKEYS = ();
    $CBB::sorted_catkeys = 0;

    return "ok";
}


# set @CBB::CATKEYS = sorted list of transaction keys
sub sort_catkeys {
    $CBB::sorted_catkeys = 1;

    print DEBUG "sort_catkeys()\n" if $CBB::debug;
    @CBB::CATKEYS = sort(keys %CBB::CATS);
}


# edit a category in the category list
sub edit_cat {
    # in: category
    # out: category

    my($cat) = @_;
    my($key, $desc, $tax) = split(/\t/, $cat);

    $CBB::sorted_catkeys = 0;

    # adds new entry if non-existent
    $CBB::CATS{$key} = "$desc\t$tax";

    print DEBUG "cat-edit:  $cat\n" if $CBB::debug;

    return "$cat";
}


# insert a category into the category list
sub insert_cat {
    # in: category
    # out: category

    my($cat) = @_;
    my($key, $desc, $tax) = split(/\t/, $cat);

    $CBB::sorted_catkeys = 0;

    $CBB::CATS{$key} = "$desc\t$tax";
    push( @CBB::CATKEYS, $key );

    print DEBUG "cat-insert:  $cat\n" if $CBB::debug;

    return "$cat";
}

# insert an into the category list
sub insert_cat_account {
    # in: group accountname desc
    # out: category

    my($cat) = @_;
    my($group, $acct, $desc) = split(/\t/, $cat);
    my $key = "[$group.$acct]";

    $CBB::sorted_catkeys = 0;

    $CBB::CATS{$key} = "$desc";
    push( @CBB::CATKEYS, $key );

    print DEBUG "cat-insert:  $key\n" if $CBB::debug;

    return "$cat";
}

# delete a category from the category list
sub delete_cat {
    # in: category

    my($cat) = @_;
    my($key, $desc, $tax) = split(/\t/, $cat);

    $CBB::sorted_catkeys = 0;

    delete $CBB::CATS{$key};
    for( my $i=0; $i<=$#CBB::CATKEYS; ++$i ) {
      if( "$CBB::CATKEYS[$i]" eq "$key" ) {
        splice( @CBB::CATKEYS, $i, 1 );
        last;
      }
    }

    print DEBUG "cat-deleted:  $cat\n" if $CBB::debug;

    return "$cat";
}

# delete an account entry from the category list
sub delete_cat_account {
    # in: group, accountname 

    my($cat) = @_;
    my($group, $acct) = split(/\t/, $cat);
    my $key = "[$group.$acct]";

    $CBB::sorted_catkeys = 0;

    delete $CBB::CATS{"$key"};
    for( my $i=0; $i<=$#CBB::CATKEYS; ++$i ) {
      if( "$CBB::CATKEYS[$i]" eq "$key" ) {
        splice( @CBB::CATKEYS, $i, 1 );
        last;
      }
    }

    print DEBUG "cat-deleted:  $key\n" if $CBB::debug;

    return "$cat";
}



# attempt to find a category matching the key
# incomplete keys are allowed
# returns 2 lines: 
#  <keyword> match match match
#  VALUE <Nth match>
# KEYWORD: UNIQUE PARTIAL COMPLET NONE
sub find_cat {
    # in: N<tab>key
    # out: category

    my($nkey, $key) = split ('\t', $_[0]);
    my(@result, $words, $count, $i, $match, $catkey);

    if ($CBB::sorted_catkeys == 0) {
	&sort_catkeys();
    }

    if ( $key ne "" ) {
	# escape any '[' and ']' in $key
	$key =~ s/\[/\\\[/g;
	$key =~ s/\]/\\\]/g;
	print DEBUG "$key\n" if $CBB::debug;
	$count = 0;
	$match = 0;
	@result = ();

        foreach $catkey (@CBB::CATKEYS) {
	    if ( $catkey =~ m/^$key/i ) {
		print DEBUG "$catkey <=> $key\n" if $CBB::debug;
                $count++;

		if ($catkey =~ m/^$key$/i) {
		    print DEBUG "exact match\n" if $CBB::debug;
		    $match = 1;
		}
		@result = (@result, $catkey);
	    }
	}
	if ($count && ($nkey < $count)) {
	    for ($i = 0; $i < @result; $i++) {
		$words = $words . " " . $result[$i];
	    }
	    if (($count == 1) && $match) { 
		return "EXACT\t$words\nVALUE\t$result[$nkey]";
	    } 
	    if ($count == 1) { 
		return "UNIQUE\t$words\nVALUE\t$result[$nkey]";
	    } 
	    if ( $match ) {
		return "COMPLETE\t$words\nVALUE\t$result[$nkey]";
	    } 
	    return "PARTIAL\t$words\nVALUE\t$result[$nkey]";
	}
    }

    print DEBUG "found none\n" if $CBB::debug;
    return "NONE\t \n";
}

# returns the entire category list in one big chunk.
sub all_cats {
    # out: category list

    my($key);

    $| = 0;				# turn off buffer flushing

    if ($CBB::sorted_catkeys == 0) {
	&sort_catkeys();
    }

    foreach $key (@CBB::CATKEYS) {
	print ("$key\t$CBB::CATS{$key}\n");
    }

    $| = 1;				# turn buffer flushing back on

    return "none";
}

# each [word] catkey should correspond to an account key
# returns 'ok' if all exist, returns 'error' if any missing
sub check_cats_accounts {

    my($catkey, $acct);

    if ($CBB::sorted_catkeys == 0) { &sort_catkeys(); }

    foreach $catkey (@CBB::CATKEYS) {
	if ( $catkey =~ m/^\[.*\]$/ ) {
	    $acct = $catkey;
	    $acct =~ s/^\[//;
	    $acct =~ s/\]$//;
            # Accounts are stored in group.acctname format...
            $acct = (split( /\./, $acct ))[1];
	    # print STDERR "catkey: $catkey, acct: $acct\n";
	    if ($CBB::ACCTS{$acct} eq "") {
		return "error";
	    }
	}
    }

    return "ok";
}

# each [word] catkey should correspond to an account key
# returns list of missing accounts, ending with "none"
sub list_missing_accounts {

    my($catkey, $acct);

    if ($CBB::sorted_catkeys == 0) { &sort_catkeys(); }

    foreach $catkey (@CBB::CATKEYS) {
	if ( $catkey =~ m/^\[.*\]$/ ) {
	    $acct = $catkey;
	    $acct =~ s/^\[//;
	    $acct =~ s/\]$//;
	    # print STDERR "catkey: $catkey, acct: $acct\n";
	    if ($CBB::ACCTS{$acct} eq "") {
		print "$acct\n";
	    }
	}
    }

    return "none";
}

sub def_cats {
    # in: nothing
    # out: result

    if ($CBB::categories eq "") { return "error"; }

    return ($CBB::categories);
}

# load a categories list
sub load_cats {
    # in: root pathname, category filename
    # out: result

    my($rootPath, $name) = split( /\t+/, $_[0] );
    my($group);

    $CBB::sorted_catkeys = 0;

    my $category = $name;
    my $pathname = "$rootPath/$CBB::group";
    my $filename = "$pathname/$category.cat";
    return "error" if ( !stat $filename );
    open(LOADCATS, "<$filename") || return "error";

    while ( <LOADCATS> ) {
	chop;
	if ( m/^#/ ) {
	     if ( m/^#\s*GROUP\*=/ ) {
		  ($group) = /^#\s*GROUP\*=\s*(.*)$/;
		  print STDERR "group: $group\n";
		  if ($group ne $CBB::group) {
		      print DEBUG "category group doesn't match\n" if $CBB::debug;
		      return "error";
		  }
	      }
	 } else {
	     if ( ! m/\t/ ) { s/:/\t/g; }
	     &insert_cat ($_);
	 }
    }

    # after loading cats, we should check constistency with accounts:
    # each [word] entry should correspond to an account entry

    sort_catkeys ();
    if (!check_cats_accounts ()) { return "error"; }

    close(LOADCATS);

    return "ok";
}

# save the category list
# ASSUMPTION:  $CBB::categories already set!
sub save_cats {
    # in: root pathname
    # out: result

    my($key);
    my $pathname = shift @_;
    $pathname .= "/$CBB::group";

    print DEBUG "save_cats to file $pathname/$CBB::categories\n" if $CBB::debug;

    if ($CBB::categories eq "") { return "error"; }
    if ($CBB::sorted_catkeys == 0) { &sort_catkeys(); }

    # Does category file exist?
    my $filename = "$pathname/$CBB::categories.cat";
    open(SAVECATS, ">$filename") || return "error";

    print SAVECATS "# CBB-Category File\n";
    print SAVECATS "#\n";
    print SAVECATS "# GROUP = $CBB::group\n";
    print SAVECATS "# values on each line are:\n";
    print SAVECATS "# (key) (name) (tax)\n";
    print SAVECATS "#\n";
    
    foreach $key (@CBB::CATKEYS) {
        print( SAVECATS "$key\t$CBB::CATS{$key}\n" );
    }

    close(SAVECATS);

    return "ok";
}

# make a category file
sub make_cats {
    # in: filename (fully-qualified)
    # out: result
    
    my($rootPath, $name) = split( /\t+/, $_[0] );

    print DEBUG "Make cats $name\n" if $CBB::debug;

    init_cats ();
    
    $CBB::categories = $name;

    save_cats ( $rootPath );
    save_accts ();

    return "ok";
}

&init_cats();
